classdef Result < handle
    % RESULT Result obtained from a simulation with Rotare.
    %   This class defines the Result object that can be used to store the solution an entire Rotare
    %   simulation.
    %
    % Notes:
    %   - Almost all properties are private to prevent accidental overwriting.
    %   - Most properties are structured as [nRotor, nOperPoint]
    % -----
    %
    % Result properties:
    %   leishman - Results obtained with the leishman solver
    %   indfact  - Results obtained with the indFact solver
    %   indvel   - Results obtained with the indVel solver
    %   stahlhut - Results obtained with the stalhlut solver
    %   operPts  - Table with the operating points for each individual result
    %   ModExt   - Modelling extension options (losses, etc)
    %   cT       - Table with all Thrust coefficients
    %   cP       - Table with all Power coefficients
    %   cQ       - Table with all Torque coefficients
    %
    % Result methods:
    %   save      - Save the Result object to file
    %   summarize - Output a summary table with the operating points and the performance
    %   filter    - Filter for a specific operating point
    %   plotperf  - Plot and compare rotor performances
    %
    % Result constructor:
    %   Result = Result() creates an empty object.
    %
    % Constructor inputs:
    %
    % See also: Rotor, rotare, template, af_tools.Airfoil.
    %
    % <a href="https:/gitlab.uliege.be/rotare/documentation">Complete documentation (online)</a>

    % ----------------------------------------------------------------------------------------------
    % FIXME: Make plot methods to compare two results easily
    % ----------------------------------------------------------------------------------------------
    % (c) Copyright 2022-2023 University of Liege
    % Author: Thomas Lambert <t.lambert@uliege.be>
    % ULiege - Aeroelasticity and Experimental Aerodynamics
    % MIT License
    % Repo: https://gitlab.uliege.be/rotare/rotare
    % Docs: https://gitlab.uliege.be/rotare/documentation
    % Issues: https://gitlab.uliege.be/rotare/rotare/-/issues
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    % Base properties of the elements
    properties
        leishman (:, :) OperRotor  % Results obtained with the leishman solver
        indfact (:, :) OperRotor   % Results obtained with the indFact solver
        indvel (:, :) OperRotor    % Results obtained with the indVel solver
        stahlhut (:, :) OperRotor  % Results obtained with the stalhlut solver
        operPts table              % Table of operating points for each individual result
        ModExt (1, 1) struct       % Modelling enstension options (losses, etc)
    end

    properties (Dependent)
        cT
        cP
        cQ
    end

    properties (Constant, Hidden)
        SOLVERS = {'leishman', 'indfact', 'indvel', 'stahlhut'} % List of allowed solvers
    end

    methods

        function self = Result()
            % Result Constructor.
            %   Creates the Result object.
            %
            % Currently empty on purpose. Results are to be assigned manually.
        end

        % ---------------------------------------------
        % Set methods
        function set.operPts(self, val)
            % Note: A bug in Matlab prevents the use of variable names when size checks are added
            % in the property definition. See: https://stackoverflow.com/q/48423003
            self.operPts = val;

            % Properly name and group operating points for multi rotors
            % Must merge in that order to have correct indexes
            nRots = (size(val, 2) - 2) / 2;
            self.operPts = mergevars(self.operPts, 3 + nRots:size(val, 2));
            self.operPts = mergevars(self.operPts, 3:3 + nRots - 1);

            self.operPts.Properties.VariableNames = {'altitude', 'speed', 'rpm', 'collective'};

        end

        % ---------------------------------------------
        % Get methods for dependent properties
        function cT = get.cT(self)
            cT = maketable(self, 'cT');
        end

        function cQ = get.cQ(self)
            cQ = maketable(self, 'cQ');
        end

        function cP = get.cP(self)
            cP = maketable(self, 'cP');
        end

        % ---------------------------------------------
        % Other methods
        save(self, SaveOpts)  % Save the results to file
        summaryTable = summarize(self, propList) % Summarize results
        [filtered, idx] = filter(self, alt, speed, rpm, coll) % Filter results
    end
end

function resTable = maketable(self, prop)
    % MAKETABLE Gather properties from all solvers and all operating points in a table

    resTable = table('Size', [size(self.operPts, 1), 0]);

    for iSolv = 1:length(self.SOLVERS)
        thissolver = self.SOLVERS{iSolv};
        tmpTable = [];

        for iRotor = size(self.(thissolver), 2):-1:1
            for iOp = size(self.operPts, 1):-1:1
                tmpTable(iOp, iRotor) = self.(thissolver)(iOp, iRotor).(prop);
            end
        end

        if ~isempty(tmpTable)
            resTable = addvars(resTable, tmpTable, 'NewVariableNames', [prop, '_', thissolver]);
        end
    end

end
